/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2011 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "generic_animations.h"
#include "dbus.h"
#include "anidata_p.h"

#include <kwinconfig.h>
#include <kconfiggroup.h>
#include <kwindowsystem.h>

#include <QDBusConnection>
#include <QTimerEvent>

#include <QtDebug>

using namespace KWin;
KWIN_EFFECT( generic_animations, GenericAnimations )

static Atom slideAtom;

#include "animation.defaults"

GenericAnimations::GenericAnimations() : AnimationEffect()
, m_lastActive(0)
, m_moveResizeWindow(0)
, m_implicitActivationCandidate(0)
, m_resizeConnected(false)
, m_implicitActivationResetTimer(0)
, m_activationTimer(0)
{
    initDefaultAnimations();

    slideAtom = XInternAtom( display(), "_KDE_SLIDE", False );
    effects->registerPropertyType( slideAtom, true );
    // TODO hackish way to announce support, make better after 4.0
    unsigned char dummy = 0;
    XChangeProperty(display(), rootWindow(), slideAtom, slideAtom, 8, PropModeReplace, &dummy, 1);

    QDBusConnection::sessionBus().registerObject("/GenericAnimations", this);
    new GenericAnimationsAdaptor(this);
#if OLD_SIGNATURE
    connect( effects, SIGNAL(windowClosed(EffectWindow*)), SLOT(windowClosed(EffectWindow*)) );
#else
    connect( effects, SIGNAL(windowClosed(KWin::EffectWindow*)), SLOT(windowClosed(KWin::EffectWindow*)) );
#endif

    reconfigure( ReconfigureAll );
}

GenericAnimations::~GenericAnimations()
{
    effects->registerPropertyType( slideAtom, false );
}

void GenericAnimations::reconfigure( ReconfigureFlags )
{
    KConfigGroup conf = effects->effectConfig( "GenericAnimations" );
#if OLD_SIGNATURE
    effects->disconnect( SIGNAL(windowActivated(EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowAdded(EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowMinimized(EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowUnminimized(EffectWindow*)), this );
    effects->disconnect( SIGNAL(desktopChanged(int,int)), this );
//     effects->disconnect( SIGNAL(desktopChanged(int,int,EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowOpacityChanged(EffectWindow*,qreal, qreal)), this );
    effects->disconnect( SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)), this );
//     effects->disconnect( SIGNAL(propertyNotify(EffectWindow*,long)), this );
#else
    effects->disconnect( SIGNAL(windowActivated(KWin::EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowAdded(KWin::EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowMinimized(KWin::EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowUnminimized(KWin::EffectWindow*)), this );
    effects->disconnect( SIGNAL(desktopChanged(int,int)), this );
    effects->disconnect( SIGNAL(desktopChanged(int,int,KWin::EffectWindow*)), this );
    effects->disconnect( SIGNAL(windowOpacityChanged(KWin::EffectWindow*,qreal, qreal)), this );
    effects->disconnect( SIGNAL(windowGeometryShapeChanged(KWin::EffectWindow*, const QRect&)), this );
//     effects->disconnect( SIGNAL(propertyNotify(KWin::EffectWindow*,long)), this );
#endif
    m_resizeConnected = false;

    m_animations.clear();

    QList<AniData> list;
    QStringList keys = conf.keyList();
    foreach (const QString &key, defaultAnimations.keys())
        if (!keys.contains(key))
            keys << key;
    foreach (const QString &key, keys) {
        if (key.endsWith("Animations")) {
            const QString str = conf.readEntry(key, defaultAnimations.value(key, QString()));
            list = AniData::list(str);
            if (!list.isEmpty())
                m_animations.insert(key.left(key.size() - 10), list);
        }
    }
#if OLD_SIGNATURE
    if ((m_animateMaximize = conf.readEntry("UnMaximize", true)))
    {
        connect( effects, SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)),
                           SLOT(windowGeometryShapeChanged( EffectWindow*, const QRect&)) );
        m_resizeConnected = true;
        connect( effects, SIGNAL(windowStartUserMovedResized(EffectWindow*)), SLOT(moveResizeStarted(EffectWindow*)) );
        connect( effects, SIGNAL(windowFinishUserMovedResized(EffectWindow*)), SLOT(moveResizeStopped(EffectWindow*)) );
    }
    if (conf.readEntry("Opacity", true))
        connect( effects, SIGNAL(windowOpacityChanged(EffectWindow*,qreal, qreal)), SLOT(opacityChanged(EffectWindow*,qreal, qreal)) );
    m_slidePopups = conf.readEntry("SlidingPopups", true);
//     if (m_slidePopups)
//         connect(effects, SIGNAL(propertyNotify(EffectWindow*,long)), this, SLOT(propertyChanged(EffectWindow*,long)));

    if (m_slidePopups || m_animations.contains("Add"))
        connect( effects, SIGNAL(windowAdded(EffectWindow*)), SLOT(windowAdded(EffectWindow*)) );
    // "Close" is _always_ connected in constructor to hook on _before_ the baseclass!
    if (m_animations.contains("Minimize") || m_animations.contains("ShowDesktop"))
        connect( effects, SIGNAL(windowMinimized(EffectWindow*)), SLOT(windowMinimized(EffectWindow*)) );
    if (m_animations.contains("Unminimize") || m_animations.contains("HideDesktop"))
        connect( effects, SIGNAL(windowUnminimized(EffectWindow*)), SLOT(windowUnminimized(EffectWindow*)) );

    if (m_animations.contains("LeavingDesktop") || m_animations.contains("EnteringDesktop") ||
        m_animations.contains("LeavingDesktopReverse") || m_animations.contains("EnteringDesktopReverse"))
    {
        if ( !connect( effects, SIGNAL(desktopChanged(int,int, EffectWindow*)), SLOT(desktopChanged(int,int, EffectWindow*)) ) )
            connect( effects, SIGNAL(desktopChanged(int,int)), SLOT(desktopChanged(int,int)) );

        if (!m_animateMaximize) // otherwise is already connected - TODO: do we need a flag to setup connections?
        {   // to avoid actions while dragging windows across desktops
            connect( effects, SIGNAL(windowStartUserMovedResized(EffectWindow*)), SLOT(moveResizeStarted(EffectWindow*)) );
            connect( effects, SIGNAL(windowFinishUserMovedResized(EffectWindow*)), SLOT(moveResizeStopped(EffectWindow*)) );
        }
    }
    if (m_animations.contains("Activation") || m_animations.contains("ImplicitActivation") || m_animations.contains("Deactivation"))
        connect( effects, SIGNAL(windowActivated(EffectWindow*)), SLOT(windowActivated(EffectWindow*)) );
#else
if ((m_animateMaximize = conf.readEntry("UnMaximize", true)))
    {
        connect( effects, SIGNAL(windowGeometryShapeChanged(KWin::EffectWindow*, const QRect&)),
                           SLOT(windowGeometryShapeChanged( KWin::EffectWindow*, const QRect&)) );
        m_resizeConnected = true;
        connect( effects, SIGNAL(windowStartUserMovedResized(KWin::EffectWindow*)), SLOT(moveResizeStarted(KWin::EffectWindow*)) );
        connect( effects, SIGNAL(windowFinishUserMovedResized(KWin::EffectWindow*)), SLOT(moveResizeStopped(KWin::EffectWindow*)) );
    }
    if (conf.readEntry("Opacity", true))
        connect( effects, SIGNAL(windowOpacityChanged(KWin::EffectWindow*,qreal, qreal)),
                            SLOT(opacityChanged(KWin::EffectWindow*,qreal, qreal)) );
    m_slidePopups = conf.readEntry("SlidingPopups", true);
//     if (m_slidePopups)
//         connect(effects, SIGNAL(propertyNotify(EffectWindow*,long)), this, SLOT(propertyChanged(EffectWindow*,long)));

    if (m_slidePopups || m_animations.contains("Add"))
        connect( effects, SIGNAL(windowAdded(KWin::EffectWindow*)), SLOT(windowAdded(KWin::EffectWindow*)) );
    // "Close" is _always_ connected in constructor to hook on _before_ the baseclass!
    if (m_animations.contains("Minimize") || m_animations.contains("ShowDesktop"))
        connect( effects, SIGNAL(windowMinimized(KWin::EffectWindow*)), SLOT(windowMinimized(KWin::EffectWindow*)) );
    if (m_animations.contains("Unminimize") || m_animations.contains("HideDesktop"))
        connect( effects, SIGNAL(windowUnminimized(KWin::EffectWindow*)), SLOT(windowUnminimized(KWin::EffectWindow*)) );

    if (m_animations.contains("LeavingDesktop") || m_animations.contains("EnteringDesktop") ||
        m_animations.contains("LeavingDesktopReverse") || m_animations.contains("EnteringDesktopReverse"))
    {
        if ( !connect( effects, SIGNAL(desktopChanged(int,int, KWin::EffectWindow*)), SLOT(desktopChanged(int,int, KWin::EffectWindow*)) ) )
            connect( effects, SIGNAL(desktopChanged(int,int)), SLOT(desktopChanged(int,int)) );

        if (!m_animateMaximize) // otherwise is already connected - TODO: do we need a flag to setup connections?
        {   // to avoid actions while dragging windows across desktops
            connect( effects, SIGNAL(windowStartUserMovedResized(KWin::EffectWindow*)), SLOT(moveResizeStarted(KWin::EffectWindow*)) );
            connect( effects, SIGNAL(windowFinishUserMovedResized(KWin::EffectWindow*)), SLOT(moveResizeStopped(KWin::EffectWindow*)) );
        }
    }
    if (m_animations.contains("Activation") || m_animations.contains("ImplicitActivation") || m_animations.contains("Deactivation"))
        connect( effects, SIGNAL(windowActivated(KWin::EffectWindow*)), SLOT(windowActivated(KWin::EffectWindow*)) );
#endif

}


static bool matches( EffectWindow *w, NET::WindowTypeMask type)
{
//     return w->windowType() != NET::Unknown && (1<<w->windowType()) & type;
    if ( (type & NET::NormalMask) && w->isNormalWindow() )
        return true;
    if ( (type & NET::PopupMenuMask) && w->isPopupMenu() )
        return true;
    if ( (type & NET::DialogMask) && w->isDialog() )
        return true;
    if ( (type & NET::MenuMask) && w->isMenu() )
        return true;
    if ( (type & NET::TooltipMask) && w->isTooltip() )
        return true;
    if ( (type & NET::DropdownMenuMask) && w->isDropdownMenu() )
        return true;
    if ( (type & NET::ComboBoxMask) && w->isComboBox() )
        return true;
    if ( (type & NET::DNDIconMask) && w->isDNDIcon() )
        return true;
    if ( (type & NET::UtilityMask) && w->isUtility() )
        return true;
    if ( (type & NET::SplashMask) && w->isSplash() )
        return true;
    if ( (type & NET::DockMask) && w->isDock() )
        return true;
    if ( (type & NET::ToolbarMask) && w->isToolbar() )
        return true;
    if ( (type & NET::NotificationMask) && w->isNotification() )
        return true;
    if ( (type & NET::DesktopMask) && w->isDesktop() )
        return true;
    if ( (type & NET::NotificationMask) && w->isNotification() )
        return true;

    return false;
// NET::Unknown
// NET::Override
// NET::TopMenu

// isManaged()
// isModal()
// isSpecialWindow()

}

static bool ignoreShowHide(EffectWindow *w, bool show)
{
    // doesn't work - the proerperty is there, but not set to anything atm
    // static Atom screenlocker = XInternAtom(QX11Info::display(), "_KDE_SCREEN_LOCKER", False);
    // !w->readProperty(screenlocker, screenlocker, 32).isEmpty();

    
    const QString wc(w->windowClass());
     // "ksplashx ksplashx" "ksplashsimple ksplashsimple" "qt-subapplication ksplashqml" "kscreenlocker_greet"
    return wc.contains("ksplash") || (show && wc.startsWith("kscreenlocker"));
}

static int mouseToAnchor(EffectWindow *w, int m, AnimationEffect::Attribute a)
{
    if ( !(m & AnimationEffect::Mouse) || (a != AnimationEffect::Scale && a != AnimationEffect::Position && a != AnimationEffect::Size) )
        return m;

    QRect geo = w->geometry();
    QPoint cursor = QCursor::pos();
    m = 0;
    if ( cursor.x() < geo.x() + geo.width()/4 )
        m |= AnimationEffect::Left;
    else if ( cursor.x() > geo.x() + 3*geo.width()/4 )
        m |= AnimationEffect::Right;
    if ( cursor.y() < geo.y() + geo.height()/4 )
        m |= AnimationEffect::Top;
    else if ( cursor.y() > geo.y() + 3*geo.height()/4 )
        m |= AnimationEffect::Bottom;

    return m;
}

void GenericAnimations::animationEnded(EffectWindow *w, Attribute a, uint meta)
{
    if (meta & (1<<24)) {
        PendingResizes::iterator it = m_pendingResizes.find(w);
        if (it != m_pendingResizes.end()) { // supposed to be true
            it->removeFirst();
            if (it->isEmpty()) {
                m_pendingResizes.erase(it);
                return;
            }
            animateResize(w, it->first().first, it->first().second);
            return;
        }
    }

    if (a != Position)
        return;

    int i = m_slidingPopups.indexOf( w );
    if (i < 0)
        return;
    m_slidingPopups.removeAt(i);
    if (m_slidingPopups.isEmpty()) {
        EffectWindowList windows = effects->stackingOrder();
        foreach (EffectWindow *dock, windows) {
            if (dock->isDock())
                effects->setElevatedWindow(dock, false);
        }
    }
}

void GenericAnimations::opacityChanged( EffectWindow *w, qreal oldOpacity, qreal newOpacity )
{
    float diff = qAbs(newOpacity - oldOpacity);
    if (diff > 0.05)
        animate( w, Opacity, 0, qMax(0.0f,1-(1-diff)*(1-diff))*500, FPx2(1.0), QEasingCurve::Linear, 0, FPx2(oldOpacity/newOpacity) );
}

// void GenericAnimations::propertyChanged(EffectWindow *w, long atom)
// {
//     if (w && (Atom)atom == slideAtom)
//         slideInOutPopup(w, !w->isDeleted());
// }

#define IF_HAVE_ANIMATION(_ANI_) AniHash::const_iterator animation = m_animations.find(_ANI_); if (animation != m_animations.constEnd())
#define META mouseToAnchor(w, d.meta, d.attribute)

void GenericAnimations::trigger(QString key, WId id)
{
    IF_HAVE_ANIMATION(key) {
        EffectWindowList windows;
        if (id) {
            if (EffectWindow *w = effects->findWindow(id))
                windows << w;
            else
                return; // there's no such window
        }
        else
            windows = effects->stackingOrder();
        foreach (EffectWindow *w, windows) {
            foreach (const AniData &d, *animation)
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
}


void GenericAnimations::timerEvent(QTimerEvent *te)
{
    if (te->timerId() == m_implicitActivationResetTimer)
    {
        killTimer(m_implicitActivationResetTimer);
        m_implicitActivationResetTimer = 0;
        m_implicitActivationCandidate = 0;
    } else if (te->timerId() == m_activationTimer) {
        killTimer(m_activationTimer);
        m_activationTimer = 0;
        if (m_lastActive) {
            QString reason("Activation");
            if (m_lastActive == m_implicitActivationCandidate)
            {
                m_implicitActivationCandidate = 0;
                reason = "ImplicitActivation";
            }
            IF_HAVE_ANIMATION(reason) {
                foreach (const AniData &d, *animation)
                    if (matches(m_lastActive, d.windowType))
                        animate(m_lastActive, d.attribute, mouseToAnchor(m_lastActive, d.meta, d.attribute), d.duration, d.to, d.curve, d.time, d.from);
            }
        }
    }
    else
        AnimationEffect::timerEvent(te);
}

static EffectWindow *slidingWindow = 0;
void GenericAnimations::desktopChanged(int oldDesktop, int newDesktop, EffectWindow *with)
{
    slidingWindow = with;
    desktopChanged(oldDesktop, newDesktop);
    slidingWindow = 0;
}

void GenericAnimations::desktopChanged(int oldDesktop, int newDesktop)
{
    if (m_moveResizeWindow)
        return;
    EffectWindowList allWindows = effects->stackingOrder();
    AniHash::const_iterator animation, leaving, entering;
    int d1 = newDesktop - oldDesktop, d2 = oldDesktop - newDesktop;
    if (d1 < 0) d1 += effects->numberOfDesktops();
    if (d2 < 0) d2 += effects->numberOfDesktops();
    if (d1 > d2) {
        leaving = m_animations.find("LeavingDesktopReverse");
        entering = m_animations.find("EnteringDesktopReverse");
    } else {
        leaving = m_animations.find("LeavingDesktop");
        entering = m_animations.find("EnteringDesktop");
    }
    foreach (EffectWindow *w, allWindows) {
        if (w->isMinimized() || w == slidingWindow)
            continue;
        const bool onOld = w->isOnDesktop(oldDesktop), onNew = w->isOnDesktop(newDesktop);
        if (onOld && !onNew)
            animation = leaving;
        else if (onNew && !onOld)
            animation = entering;
        else
            continue;
        if (animation != m_animations.constEnd()) {
            foreach (const AniData &d, *animation)
                if ( matches( w, d.windowType) )
                    animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
}

void GenericAnimations::moveResizeStarted(EffectWindow *w)
{
    m_moveResizeWindow = w;
    if (m_animateMaximize) {
#if OLD_SIGNATURE
        effects->disconnect( SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)), this );
#else
    effects->disconnect( SIGNAL(windowGeometryShapeChanged(KWin::EffectWindow*, const QRect&)), this );
#endif
        m_resizeConnected = false;
    }
}

void GenericAnimations::moveResizeStopped(EffectWindow *w)
{
    Q_UNUSED(w);
    m_moveResizeWindow = 0;
    if (m_animateMaximize && !m_resizeConnected) {
#if OLD_SIGNATURE
        connect( effects, SIGNAL(windowGeometryShapeChanged(EffectWindow*, const QRect&)),
                          SLOT(windowGeometryShapeChanged( EffectWindow*, const QRect&)) );
#else
        connect( effects, SIGNAL(windowGeometryShapeChanged(KWin::EffectWindow*, const QRect&)),
                          SLOT(windowGeometryShapeChanged(KWin::EffectWindow*, const QRect&)) );
#endif
        m_resizeConnected = true;
    }
}

static const int size_of_long = sizeof(long);
bool GenericAnimations::slideInOutPopup(EffectWindow *w, bool in)
{
    if (!m_slidePopups)
        return false;

    QByteArray slideData = w->readProperty( slideAtom, slideAtom, 32 );
    if (slideData.size() < 2*size_of_long)
        return false;

    const long *ld = reinterpret_cast<const long*>(slideData.constData());
    // "start = ld[ 0 ];"

    uint a;
    FPx2 other;
    const QRect geo(w->geometry());
    switch (ld[1]) {
    case 0: // slide from/to left
        a = Right|Top;
        other.set(0, geo.y());
        break;
    case 1:
        a = Bottom|Left;
        other.set(geo.x(), 0);
        break; // slide from/to top
    case 2:
        a = Left|Top;
        other.set(effects->clientArea(ScreenArea , w).right(), geo.y());
        break; // slide from/to right
    case 3:
        a = Top|Left;
        other.set(geo.x(), effects->clientArea(ScreenArea , w).bottom());
        break; // slide from/to bottom
    default:
        return false;
    }

    uint meta(0);
    setMetaData(in ? SourceAnchor : TargetAnchor, a, meta);

    long time = 300;
    if (!in && slideData.size() >= 4*size_of_long)
        time = ld[3];
    else if (slideData.size() >= 3*size_of_long)
        time = ld[2];

    // yes - append blindly! "if (!m_slidingPopups.contains(w))" is wrong because we can trigger one animation while the other is running
    // the first exiting would than redo elevations though in fact this window still needs one.
    m_slidingPopups << w;

    if (in) {
        animate( w, Position, meta, time, FPx2(), QEasingCurve::OutQuart, 0, other );
        animate( w, Clip, meta, time, FPx2() );
        // NOTICE, TODO "better"
        // because clipping atm. is only supported via the repaint region (ie. clipping does not compensate translation etc.)
        // we "fake" a clipping "animation" by keeping the clipregion actually static during the animation
//         animate( w, Clip, meta, time, FPx2(), QEasingCurve::OutQuart, 0, FPx2(ld[1]%2,(ld[1]+1)%2) );
    }
    else {
        animate( w, Position, meta, time, other, QEasingCurve::InQuart );
        animate( w, Clip, meta, time, FPx2() );
        // NOTICE, TODO see above
//         animate( w, Clip, meta, time, FPx2(ld[1]%2,(ld[1]+1)%2), QEasingCurve::InQuart );
    }

    return true;
}

#define RESTART(_T_, _TT_) if (_T_) killTimer(_T_); _T_ = startTimer(_TT_)

void GenericAnimations::windowActivated( EffectWindow *w )
{
    if (m_lastActive && !(m_lastActive->isMinimized() || m_lastActive->isDeleted())) {
        IF_HAVE_ANIMATION("Deactivation") {
            foreach (const AniData &d, *animation)
                if ( matches( m_lastActive, d.windowType) )
                    animate( m_lastActive, d.attribute, mouseToAnchor(m_lastActive, d.meta, d.attribute), d.duration, d.to, d.curve, d.time, d.from );
        }
    }

    m_lastActive = w;

    RESTART(m_activationTimer, 75);
}


static bool showingDesktop()
{
    static const unsigned long props[2] = {0, NET::WM2ShowingDesktop};
    return NETRootInfo(QX11Info::display(),  props, 2).showingDesktop();
}

void GenericAnimations::windowAdded( EffectWindow* w )
{
    if ( w->isMinimized() || w->isDeleted() || showingDesktop() || !w->isOnDesktop(effects->currentDesktop()))
        return;

    if (ignoreShowHide(w, true))
        return;

    if (!slideInOutPopup(w, true)) {
        IF_HAVE_ANIMATION("Add") {
            foreach (const AniData &d, *animation)
                if ( matches( w, d.windowType) )
                    animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
        }
    }
    m_implicitActivationCandidate = w;
    RESTART(m_implicitActivationResetTimer, 250);
}

void GenericAnimations::windowClosed( EffectWindow* w )
{
    if (slideInOutPopup(w, false))
        return;
    if (ignoreShowHide(w, false))
        return;
    IF_HAVE_ANIMATION("Close") {
        foreach (const AniData &d, *animation)
            if ( matches( w, d.windowType) )
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
    }
}

void GenericAnimations::windowDeleted( EffectWindow* w )
{
    if (w == m_lastActive)
        m_lastActive = 0;
}


void GenericAnimations::windowGeometryShapeChanged( EffectWindow *w, const QRect &old )
{
    if (!m_animateMaximize || m_moveResizeWindow || w == m_implicitActivationCandidate)
        return;
    if (w->isDeleted() || w->isMinimized() || w->windowClass() == "kwin kwin")
        return;
    if (w->isPopupMenu())
        return; // qt does freaky things like hiding, showing and resizing the popup twice at once
    if (!old.isValid())
        return;

    const QRect geo(w->geometry());
    int d = qAbs(geo.width()-old.width()) + qAbs(geo.height()-old.height());
    if (d < 16)
        return;

    PendingResizes::iterator it = m_pendingResizes.find(w);
    if (it == m_pendingResizes.end()) {
        m_pendingResizes.insert(w, PendingResizeList() << PendingResize(old, geo));
        animateResize(w, old, geo);
    } else {
        *it << PendingResize(old, geo);
    }
}

void GenericAnimations::animateResize( EffectWindow *w, const QRect &from, const QRect &to )
{
    float time;
    const QRectF screen = effects->clientArea(ScreenArea , w);
    const QRect intersection(from & to);
    const QRect unition(from | to);
    const float fx = (unition.width() - intersection.width()) / float(screen.width());
    const float fy = (unition.height() - intersection.height()) / float(screen.height());

    // faster, but not as good as sqrt
//     time = 1.0f - qMax(fx, fy);
//     time = 1.0f - time*time*time;

    time = sqrt(qMax(fx,fy));
    time *= 400;

    animate( w, Position, (1<<24), time, FPx2(to.center()), QEasingCurve::InOutCubic, 0, FPx2(from.center()) );
    animate( w, Size, (1<<24), time, FPx2(to.size()), QEasingCurve::InOutCubic, 0, FPx2(from.size()) );
#if KDE_IS_VERSION(4,10,80)
    animate( w, CrossFadePrevious, 0, qMax(int(time), 200), FPx2(1.0) );
#endif
}

void GenericAnimations::windowMinimized( EffectWindow *w )
{
    AniHash::const_iterator animation = m_animations.find(showingDesktop() ? "ShowDesktop" : "Minimize");
    if (animation != m_animations.constEnd()) {
        foreach (const AniData &d, *animation)
            if ( matches( w, d.windowType) )
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
    }
}

void GenericAnimations::windowUnminimized( EffectWindow* w )
{
    AniHash::const_iterator animation = m_animations.find(KWindowSystem::showingDesktop() ? "HideDesktop" : "Unminimize");
    if (animation != m_animations.constEnd()) {
        foreach (const AniData &d, *animation)
            if ( matches( w, d.windowType) )
                animate( w, d.attribute, META, d.duration, d.to, d.curve, d.time, d.from );
    }
    m_implicitActivationCandidate = w;
    RESTART(m_implicitActivationResetTimer, 250);
}

#undef RESTART
#undef META

#include "moc_generic_animations.cpp"
#include "moc_dbus.cpp"
