/********************************************************************
 KWin - the KDE window manager
 This file is part of the KDE project.

Copyright (C) 2009 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "beclock_config.h"
#include <kwineffects.h>

#include <klocale.h>
#include <KActionCollection>
#include <kaction.h>
#include <kconfiggroup.h>
#include <ksystemtimezone.h>

#include <QApplication>
#include <QDesktopWidget>
#include <QVBoxLayout>


using namespace KWin;

KWIN_EFFECT_CONFIG( beclock, BeClockConfig )

BeClockConfigForm::BeClockConfigForm(QWidget* parent) : QWidget(parent)
{
    setupUi(this);
}

BeClockConfig::BeClockConfig(QWidget* parent, const QVariantList& args)
: KCModule(EffectFactory::componentData(), parent, args)
{
    QVBoxLayout* layout = new QVBoxLayout(this);
    layout->addWidget(myUi = new BeClockConfigForm(this));
    myUi->tabWidget->setCurrentIndex(0);

    connect(myUi->clockType, SIGNAL(currentIndexChanged(int)), SLOT(changed()));
    connect(myUi->shortcuts, SIGNAL(keyChange()), SLOT(changed()));
    connect(myUi->_24h, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->cfont, SIGNAL(fontSelected(const QFont&)), SLOT(changed()));
    connect(myUi->seconds, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->hideOnHover, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->date, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->dayOfWeek, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->showDigits, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->showRings, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->halign, SIGNAL(currentIndexChanged(int)), SLOT(changed()));
    connect(myUi->valign, SIGNAL(currentIndexChanged(int)), SLOT(changed()));
    connect(myUi->px, SIGNAL(valueChanged(int)), SLOT(changed()));
    connect(myUi->py, SIGNAL(valueChanged(int)), SLOT(changed()));
    connect(myUi->forceScreen, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->screen, SIGNAL(currentIndexChanged(int)), SLOT(changed()));
    connect(myUi->sz, SIGNAL(valueChanged(int)), SLOT(changed()));
    connect(myUi->background, SIGNAL(changed(const QColor&)), SLOT(changed()));
    connect(myUi->foreground, SIGNAL(changed(const QColor&)), SLOT(changed()));
    connect(myUi->mode, SIGNAL(currentIndexChanged(int)), SLOT(changed()));
    connect(myUi->topLeft, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->topRight, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->bottomLeft, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->bottomRight, SIGNAL(toggled(bool)), SLOT(changed()));

    connect(myUi->datePattern, SIGNAL(textEdited(const QString&)), SLOT(changed()));
    connect(myUi->attentionPeriod, SIGNAL(valueChanged(int)), SLOT(changed()));
    connect(myUi->countDownAttention, SIGNAL(valueChanged(int)), SLOT(changed()));
    connect(myUi->chime, SIGNAL(toggled(bool)), SLOT(changed()));
    connect(myUi->homeZone, SIGNAL(itemActivated(QTreeWidgetItem*, int)), SLOT(changed()));

    connect(myUi->showDigits, SIGNAL(toggled(bool)), SLOT(ensureOneDisplay(bool)));
    connect(myUi->showRings, SIGNAL(toggled(bool)), SLOT(ensureOneDisplay(bool)));
    connect(myUi->homeZoneFilter, SIGNAL(textChanged(const QString&)), SLOT(filterTimeZones(const QString&)));

    // Shortcut config. The shortcut belongs to the component "kwin"!
    myActionCollection = new KActionCollection( this, KComponentData("kwin") );

    myActionCollection->setConfigGroup("BeClock");
    myActionCollection->setConfigGlobal(true);

    KAction* a = (KAction*)myActionCollection->addAction( "BeClock" );
    a->setText( i18n("Toggle KWin clock effect" ));
    a->setProperty("isConfigurationAction", true);
    a->setGlobalShortcut( KShortcut( Qt::CTRL + Qt::SHIFT + Qt::Key_F12 ));

    myUi->shortcuts->addCollection(myActionCollection);

    const int ns = QApplication::desktop()->screenCount();
    for (int i=0; i<ns; ++i)
        myUi->screen->addItem( QString::number(i) );
    if ( ns == 1 )
    {
        myUi->forceScreen->hide();
        myUi->screen->hide();
    }
}

BeClockConfig::~BeClockConfig()
{
    // Undo (only) unsaved changes to global key shortcuts
    myUi->shortcuts->undoChanges();
}

void BeClockConfig::ensureOneDisplay( bool on )
{
    if (on || !sender() || myUi->showRings->isChecked() || myUi->showDigits->isChecked() )
        return;
    if ( sender() == myUi->showRings )
        myUi->showDigits->setChecked( true );
    else if ( sender() == myUi->showDigits )
        myUi->showRings->setChecked( true );
}

void BeClockConfig::filterTimeZones(const QString &filter)
{
    const int n = myUi->homeZone->topLevelItemCount();
    const int c = myUi->homeZone->columnCount();
    for (int i = 0; i < n; ++i) {
        QTreeWidgetItem *item = myUi->homeZone->topLevelItem(i);
        bool hidden = true;
        for (int j = 0; j < c; ++j) {
            if (item->text(j).contains(filter, Qt::CaseInsensitive)) {
                hidden = false;
                break;
            }
        }
        item->setHidden(hidden);
    }
}

void BeClockConfig::load()
{
    KCModule::load();
    KConfigGroup conf = EffectsHandler::effectConfig("BeClock");
    myUi->clockType->setCurrentIndex( conf.readEntry("Type", 0) );
    myUi->_24h->setChecked( conf.readEntry("24h", /*false*/ !KGlobal::locale()->use12Clock()) );
    QFont fnt;
    fnt.fromString(conf.readEntry("Font", fnt.toString()));
    myUi->cfont->setFont( fnt );
    myUi->hideOnHover->setChecked( conf.readEntry("HideOnHover", false) );
    myUi->date->setChecked( conf.readEntry("ShowDate", false) );
    myUi->dayOfWeek->setChecked( conf.readEntry("ShowDayOfWeek", true) );
    myUi->showRings->setChecked( conf.readEntry("ShowRings", true) );
    myUi->showDigits->setChecked( conf.readEntry("ShowDigits", true) );
    myUi->seconds->setChecked( conf.readEntry("ShowSeconds", false) );
    int align = conf.readEntry("Align", (int)(Qt::AlignTop|Qt::AlignRight));
    myUi->halign->setCurrentIndex( (align & Qt::AlignLeft) ? 0 : ( (align & Qt::AlignHCenter) ? 1 : 2 ) );
    myUi->valign->setCurrentIndex( (align & Qt::AlignTop) ? 0 : ( (align & Qt::AlignVCenter) ? 1 : 2 ) );
    myUi->px->setValue( conf.readEntry("X", 20) );
    myUi->py->setValue( conf.readEntry("Y", 20) );
    int screen = conf.readEntry("Screen", -1);
    if ( screen > myUi->screen->count() )
        screen = -1;
    myUi->screen->setCurrentIndex( qMax(0,screen) );
    myUi->forceScreen->setChecked( screen > -1 );
    myUi->sz->setValue( conf.readEntry("Size", 128) );
    myUi->background->setColor(conf.readEntry("Background", QColor(Qt::white)));
    myUi->foreground->setColor(conf.readEntry("Foreground", QColor(Qt::black)));
    QList<int> corners;
    corners = conf.readEntry("ElectricBorders", corners);
    if (!corners.isEmpty())
    {
        myUi->topLeft->setChecked( corners.contains(ElectricTopLeft) );
        myUi->topRight->setChecked( corners.contains(ElectricTopRight) );
        myUi->bottomLeft->setChecked( corners.contains(ElectricBottomLeft) );
        myUi->bottomRight->setChecked( corners.contains(ElectricBottomRight) );
        myUi->mode->setCurrentIndex( 1 );
    }
    myUi->datePattern->setText( conf.readEntry("DatePattern", "yyyy-MM-dd" ) );
    myUi->attentionPeriod->setValue( conf.readEntry("AttentionPeriod", 3600 ) );
    myUi->countDownAttention->setValue( conf.readEntry("CountDownAttention", 3 ) );
    myUi->chime->setChecked( conf.readEntry("Chime", false ) );

    myUi->homeZone->setSelected(conf.readEntry( "HomeZone", KSystemTimeZones::local().name() ), true);

    emit changed(false);
}

void BeClockConfig::save()
{
    KConfigGroup conf = EffectsHandler::effectConfig("BeClock");
    conf.writeEntry("Type", myUi->clockType->currentIndex());
    conf.writeEntry("24h", myUi->_24h->isChecked());
    conf.writeEntry("Font", myUi->cfont->font().toString());
    conf.writeEntry("HideOnHover", myUi->hideOnHover->isChecked());
    conf.writeEntry("ShowDate", myUi->date->isChecked());
    conf.writeEntry("ShowDayOfWeek", myUi->dayOfWeek->isChecked());
    conf.writeEntry("ShowDigits", myUi->showDigits->isChecked());
    conf.writeEntry("ShowRings", myUi->showRings->isChecked());
    conf.writeEntry("ShowSeconds", myUi->seconds->isChecked());
    conf.writeEntry("Foreground", myUi->foreground->color());
    conf.writeEntry("Background", myUi->background->color());
    int align = 0;
    int i = myUi->halign->currentIndex();
    align |= (i == 0) ? Qt::AlignLeft : (i == 1 ? Qt::AlignHCenter : Qt::AlignRight);
    i = myUi->valign->currentIndex();
    align |= (i == 0) ? Qt::AlignTop : (i == 1 ? Qt::AlignVCenter : Qt::AlignBottom);
    conf.writeEntry("Align", align);
    conf.writeEntry("X", myUi->px->value());
    conf.writeEntry("Y", myUi->py->value());
    int screen = -1;
    if ( myUi->forceScreen->isChecked() )
        screen = myUi->screen->currentIndex();
    conf.writeEntry("Screen", screen);
    conf.writeEntry("Size", myUi->sz->value());
    QList<int> corners;
    if ( myUi->mode->currentIndex() )
    {
        if ( myUi->topLeft->isChecked() )
            corners << ElectricTopLeft;
        if ( myUi->topRight->isChecked() )
            corners << ElectricTopRight;
        if ( myUi->bottomLeft->isChecked() )
            corners << ElectricBottomLeft;
        if ( myUi->bottomRight->isChecked() )
            corners << ElectricBottomRight;
    }
    conf.writeEntry("ElectricBorders", corners);

    conf.writeEntry("DatePattern", myUi->datePattern->text());
    conf.writeEntry("AttentionPeriod", myUi->attentionPeriod->value());
    conf.writeEntry("CountDownAttention", myUi->countDownAttention->value());
    conf.writeEntry("Chime", myUi->chime->isChecked());

    conf.writeEntry("HomeZone", myUi->homeZone->selection().first());

    myUi->shortcuts->save();   // undo() will restore to this state from now on
    conf.sync();
    emit changed(false);
    EffectsHandler::sendReloadMessage( "beclock" );
}

void BeClockConfig::defaults()
{
    myUi->clockType->setCurrentIndex( 0 );
    myUi->_24h->setChecked( false );
    myUi->date->setChecked( false );
    myUi->dayOfWeek->setChecked( true );
    myUi->hideOnHover->setChecked( false );
    myUi->seconds->setChecked( false );
    myUi->showDigits->setChecked( true );
    myUi->showRings->setChecked( true );
    myUi->mode->setCurrentIndex( 0 );
    myUi->halign->setCurrentIndex( 2 );
    myUi->valign->setCurrentIndex( 0 );
    myUi->px->setValue( 20 );
    myUi->py->setValue( 20 );
    myUi->screen->setCurrentIndex( 0 );
    myUi->forceScreen->setChecked( false );
    myUi->sz->setValue( 128 );
    myUi->background->setColor(Qt::white);
    myUi->foreground->setColor(Qt::black);
    myUi->shortcuts->allDefault();
    myUi->datePattern->setText( "yyyy-MM-dd" );
    myUi->attentionPeriod->setValue( 3600 );
    myUi->countDownAttention->setValue( 3 );
    myUi->chime->setChecked( false );
    myUi->homeZone->setSelected(KSystemTimeZones::local().name(), true);
    emit changed(true);
}

#include "beclock_config.moc"
